### Persephone - a socio-economic-ecological model of European agricultural landscapes.
###
### These are the tests for the core landscape functions.
###

@testset "Landscape initialisation" begin
    # initialise the landscape part of the model
    landscape = Persephone.initlandscape()
    space = GridSpace(size(landscape), periodic=false)
    properties = Dict{Symbol,Any}(:landscape=>landscape)
    model = AgentBasedModel(FarmPlot, space, properties=properties,
                            rng=Random.Xoshiro(param("core.seed")), warn=false)
    Persephone.initfields!(model)
    # these tests are specific to the Jena maps
    @test size(model.landscape) == (1754, 1602)
    @test Persephone.landcover((100,100), model) == Persephone.forest
    @test Persephone.landcover((300,1), model) == Persephone.soil
    @test Persephone.landcover((500,1), model) == Persephone.nodata
    @test Persephone.landcover((400,400), model) == Persephone.grass
    @test Persephone.landcover((800,800), model) == Persephone.agriculture
    @test Persephone.landcover((1100,1100), model) == Persephone.builtup
    @test Persephone.countfields(model) == 2092
    @test Persephone.averagefieldsize(model) == 5.37
    @test count(f -> ismissing(f.fieldid), model.landscape) == 1685573
    @test length(Persephone.farmplot((800,800), model).pixels) == 4049                
end

@testset "Event system" begin
    # initialise a basic model landscape
    landscape = Persephone.initlandscape()
    space = GridSpace(size(landscape), periodic=false)
    properties = Dict{Symbol,Any}(:date=>param("core.startdate"),
                                  :landscape=>landscape,
                                  :events=>Vector{FarmEvent}())
    model = AgentBasedModel(Union{Farmer,Animal,FarmPlot}, space, properties=properties,
                            rng=Random.Xoshiro(param("core.seed")), warn=false)
    # create some events and see whether they show up on the map and disappear as they should
    createevent!(model, [(1,1), (1,2), (1,3), (2,1), (2,3)], Persephone.tillage)
    createevent!(model, [(1,1), (1,2), (1,3), (2,2)], Persephone.sowing, 2)
    @test model.landscape[1,1].events == [Persephone.tillage, Persephone.sowing]
    @test model.landscape[2,1].events == [Persephone.tillage]
    @test model.landscape[2,2].events == [Persephone.sowing]
    stepsimulation!(model)
    @test model.landscape[1,1].events == [Persephone.sowing]
    @test model.landscape[2,1].events == []
    @test model.landscape[2,2].events == [Persephone.sowing]
    stepsimulation!(model)
    @test model.landscape[1,1].events == []
    @test model.landscape[2,1].events == []
    @test model.landscape[2,2].events == []
end

@testset "Landscape functions" begin
    # initialise a simple 6x6 test landscape
    landscape = Matrix{Pixel}(undef, 6, 6)
    for x in 1:6
        for y in 1:6
            (x in (1:2) || y in (1:4)) ? lc = Persephone.forest : lc = Persephone.grass
            landscape[x,y] = Pixel(lc, 0, [])
        end
    end
    landscape[6,4] = Pixel(Persephone.water, 0, [])
    space = GridSpace(size(landscape), periodic=false)
    properties = Dict{Symbol,Any}(:landscape=>landscape)
    model = AgentBasedModel(Animal, space, properties=properties, warn=false)
    # test the distance functions
    @test Persephone.distanceto((2,3), model, Persephone.forest) == 0
    @test Persephone.distanceto((2,3), model, Persephone.grass) == 2
    @test Persephone.distanceto((2,3), model, Persephone.water) == 4
    @test Persephone.distanceto((2,3), model, Persephone.soil) == Inf
    @test Persephone.distancetoedge((1,1), model) == 4
    @test Persephone.distancetoedge((4,4), model) == 1
    @test Persephone.distancetoedge((6,6), model) == 2
end
